Bayesian Network Learning (BNL) C++ Library
===========================================

Introduction
------------
BNL is a library for structural modeling and learning implemented in C++.
This library primarily utilizes the concept of a Bayesian network to
analyze relationships between random variables.

Dependencies
------------
To use the library in your own code, you will need:
* C++ Compiler
* [Boost](http://www.boost.org/) (Tested on 1.55.0)
* [SMILE](https://dslpitt.org/genie/)

To work on the library code, you will additionally need:
* Make build system
* [Doxygen](http://www.stack.nl/~dimitri/doxygen/)

Make sure that you have the necessary dependencies. These should be either
available on the system's PATH, or in a manually linked directory. Note that
the SMILE library is expected to be inside a "smile/" directory, e.g. the
header `<smile/smile.h>`.

If you are using the project's Makefile, the sub-directory inside the
"thirdparty/" directory corresponding to your OS will be linked during
compilation. At the time of writing, the SMILE source code is available here.

If you are using your own build system, make sure to look through this
project's Makefile for what compiler flags to use and which static libraries
to link.

Makefile Usage
--------------
The Makefile is designed to be very flexible and offers various targets. The
basic `make` command will run `make debug`, which utilizes debug mode and
turns off optimizations to help detect issues. In order to prepare an
optimized build, make sure to use the `make release` command.

When running the `make` command, any of the variables used in the Makefile
can be changed, which allows you to use different settings without manually
altering the Makefile. For example, if you write a main function in a file
test/main2.cpp, you can compile it in release mode by

    make release MAINSRC=test/main2.cpp

Some other targets include `make docs` (which will generate the Doxygen
documentation for the library), `make lib/libbnl.a` (which creates a static
library), `make clean` (which deletes files generated for the library),
and `make realclean` (which deletes everything generated).

Library Usage
-------------
The following show some example code snippets of interacting with the library.

The first shows how to load a dataset and then construct a random network
using the default settings:

    bnl::refcounted_dataset ds("dataset.txt");
    bnl::bayesian_network bn = bnl::make_random_network(ds);

The random network can then be output in various formats. One way is to use
the GeNIe format through the SMILE API:

    bn.write_file("output.xdsl");

Another way to output the network is through the free function for writing a
Graphviz file:

    std::ofstream file("output.gv");
    bnl::write_graphviz(file, bn);

One of the useful capabilities is to score structures. The score module
includes general scoring functions `bnl::log_score(...)`. One of the necessary
inputs is a local scoring function. For example, the BDeu score can be used:

    bnl::local_log_bdeu_score local_log_score;
    double score = bnl::log_score(bn, ds, local_log_score);

More details on using the library can be found on the detailed documentation
of the modules and classes.

Todo List
---------
1. Verify portability to various OS and external library versions.
2. Add other local log scoring functions
3. Convert from a Graphviz DOT file to a `bnl::bayesian_network`.
4. Add derived dataset classes for accelerated queries.
5. Implement learning algorithms, like MCMC and Greedy Hill Climbing.

Project Structure
-----------------
*Doxyfile* - configuration settings for Doxygen

*Makefile* - commands to compile the library

**bin** - executable files built by project

**build** - intermediary object files from compilation (.o)

**doc** - generated Doxygen documentation files

**include** - the application header files (.hpp)

**lib** - static library files built by project

**share** - saved network files (.xdsl) and dataset files (.txt)

**src** - the library source files (.cpp)

**test** - various test files (.cpp)

**thirdparty** - external libraries; for example, the SMILE code
