#ifndef BNL_BAYESIAN_NETWORK_HPP
#define BNL_BAYESIAN_NETWORK_HPP

#include <smile/smile.h>
#include <iostream>
#include <string>
#include <vector>

namespace bnl
{
class refcounted_dataset;
class bayesian_network : public ::DSL_network
{
	public:
/// \name (constructor)
//@{
    bayesian_network();
    bayesian_network(const ::DSL_network& net);
    bayesian_network(::DSL_network& net, bool fast);
    explicit bayesian_network(int num_nodes, int initial_delta = 50);

    /**
     * \brief Construct a Bayesian network from a file.
     *
     * A convenience constructor that allows for both initializing the network
     * object and also loading in the Bayesian network from a SMILE-readable
     * file. If there is an issue finding or parsing the file, an exception is
     * thrown.
     *
     * \param filename The name of the input network file.
     * \throw io_error if an error occurs while reading the file.
     */
    explicit bayesian_network(const ::std::string& filename);
//@}

/// \name Modifiers
//@{
    /**
     * \brief Randomize the network's arcs based on a total order.
     *
     * The maximum number of parents can be limited to reduce computation on the
     * resulting Bayesian network. In this case, there may be a bias towards
     * particular node parents in the basic algorithm, so the possible parents
     * for each node may be randomly shuffled before selection (this doesn't
     * affect the total order).
     *
     * This function always runs can_use_order to make sure that the input total
     * order is valid. This adds another O(n) to the runtime, but prevents
     * running on incorrect orders. In the case of mismatch, an exception will
     * be thrown.
     *
     * \param order The total order to test on.
     * \param max_parents The maximum number of parents for any node.
     * \param shuffle_possible_parents Whether to shuffle parents to reduce bias.
     * \throw match_error if the input total order does not match the network.
     */
    void randomize_arcs(const ::std::vector<int>& order, unsigned max_parents = 5,
                        bool shuffle_possible_parents = true);

    /**
     * \brief Randomize the network's arcs based on internal total order.
     *
     * This overload randomly generates arcs either using the randomly generated
     * total order stored inside the Bayesian network, or just uniformly-at-random
     * from all possible Bayesian networks. In the latter case, the last two
     * arguments are ignored.
     *
     * \param use_order Whether to use the underlying order or not.
     * \param max_parents The maximum number of parents for any node.
     * \param shuffle_possible_parents Whether to shuffle parents to reduce bias.
     * \throw match_error if the internal total order does not match the network.
     * \see randomize_arcs(const ::std::vector<int>&, unsigned, bool)
     * \see randomize_order
     */
    void randomize_arcs(bool use_order = true, unsigned max_parents = 5,
                        bool shuffle_possible_parents = true);

    /**
     * \brief Randomize the total order for the network.
     *
     * The total order is chosen uniformly at random from all possible orders.
     * Any SMILE error code is thrown since this means that something may be
     * fundamentally wrong inside the network.
     *
     * \throw smile_error if the nodes inside the network couldn't be accessed.
     */
    void randomize_order();
//@}

/// \name Queries
//@{
    /**
     * \brief Check that the total order can be used with this network.
     *
     * Specifically, this function checks that each node is represented inside
     * the total order exactly once.
     *
     * \param order The total order to test on.
     * \return Whether or not the network's nodes are all represented exactly
     *         once inside the total order.
     */
    bool can_use_order(const ::std::vector<int>& order) const;

    /**
     * \brief Check if the network is totally ordered by the given ordering.
     *
     * This will check that the input order can match the structure of the
     * network and verify the parent-child relationships.
     *
     * \param order The total order to test on.
     * \return Whether or not the network is totally ordered.
     * \see can_use_order
     */
    bool totally_ordered(const ::std::vector<int>& order) const;

    /**
     * \brief Check if the network is totally ordered by the internal ordering.
     * \return Whether or not the network is totally ordered.
     * \see totally_ordered(const ::std::vector<int>&) const
     */
    bool totally_ordered() const;

    /**
     * \brief Get the underlying total order. Not checked to be in sync.
     * \return The total order stored inside the network.
     */
    const ::std::vector<int>& total_order() const;

    /**
     * \brief Get the identifier corresponding to the input node handle.
     * \param handle The integer value corresponding to a particular node.
     * \return The identifier if the node exists; otherwise, an empty c-string.
     */
    const char* get_id(int handle) const;

    /**
     * \brief Get the outcome name corresponding to the node handle and index.
     *
     * In the case that the node does not exist, an empty c-string will be
     * returned. With a valid node handle, the outcome name at the given index
     * will be returned. No bounds checking is done on the index, so calling
     * with a value out-of-range causes undefined behavior.
     *
     * \param handle The integer value corresponding to a particular node.
     * \param index The index of the outcome names to access.
     * \return The identifier if the node exists; otherwise, an empty c-string.
     */
    const char* get_outcome_name(int handle, int index) const;
//@}

/// \name Input/Output
//@{
    /**
     * \brief Load in a network from the contents of a file.
     *
     * The filename should specify the location of a SMILE-parsable file as
     * described in the documentation for the DSL_network::ReadFile function.
     * The only difference is that this function will throw an appropriate
     * exception in the case of a SMILE error code.
     *
     * \param filename The name of the input network file.
     * \param filetype The format of the file (by default, it bases off extension)
     * \param ext
     * \throw io_error if an error occurs while reading the file.
     */
    void read_file(const ::std::string& filename, int filetype = 0, void* ext = 0);

    /**
     * \brief Write out an autoformatted file of the network.
     *
     * A convenience function that calls both SMILE's default-layout function
     * and write-file function. The same output can be generated using SMILE's
     * API. In the case of a SMILE error code, an exception will be thrown.
     *
     * \param filename The name of the output network file.
     * \param filetype The format of the file (by default, it bases off extension).
     * \param ext
     * \throw io_error if an error occurs while writing the file.
     */
    void write_file(const ::std::string& filename, int filetype = 0, void* ext = 0);

    /**
     * \brief Print out a node's properties.
     *
     * All of the properties of the given node handle (e.g. name, identifier,
     * conditional probabilities, outcomes, parents, and children) are inserted
     * into the given output stream.
     *
     * By default, this goes to standard output. Some minimal formatting is done
     * for a readable output. Also, an invalid handle will insert a message into
     * the stream that the node doesn't exist.
     *
     * \param handle The integer value corresponding to a particular node.
     * \param os ostream object where node properties are inserted.
     */
    void print_node_properties(int handle, ::std::ostream& os = ::std::cout) const;
    
    void probabilities(int handle);
//@}

private:
    bool randomize_arcs_without_order();

    /**
     * \brief A total order for the nodes in the network.
     *
     * This contains a generated total order of the nodes in the Bayesian
     * network. Lower index nodes precede higher index nodes; for example, if
     * the order had [...,A,B,...], then A<B, meaning that A is not a child of
     * B. It is possible for the order to fall out-of-sync with the underlying
     * network, which can be detected by bayesian_network::totally_ordered().
     */
    ::std::vector<int> total_order_;
	
};
::std::ostream& operator<<(::std::ostream& os, const bayesian_network& bn);
/**
 * \brief Generate a random network based on a dataset.
 *
 * This is mainly a convenience function that calls various available
 * functions to add nodes corresponding to variables in a dataset along with
 * randomized arcs.
 *
 * \param dataset The reference-counted dataset from which to create nodes.
 * \param use_order Whether or not to base the arcs off the random order.
 * \param max_parents The maximum number of parents for any node.
 * \param shuffle_possible_parents Whether to shuffle parents to reduce bias.
 * \throw smile_error if anything in the SMILE API returns an error code.
 * \throw match_error if detected mismatch in generated order. This should
 *        not happen normally, so it may be a sign of a more serious problem.
 * \see refcounted_dataset::prepare_structure
 * \see bayesian_network::randomize_order
 * \see bayesian_network::randomize_arcs
 */
bayesian_network add_parameters(refcounted_dataset& dataset,
                                     bool use_order = true,
                                     unsigned max_parents = 5,
                                     bool shuffle_possible_parents = true);

}
#endif // BNL_BAYESIAN_NETWORK_HPP
