/**
 * \file
 * \brief Defines the join algorithm to insert into an output stream.
 */

#ifndef BNL_ALGORITHM_JOIN_HPP
#define BNL_ALGORITHM_JOIN_HPP

#include <boost/concept_check.hpp>
#include <boost/concept/assert.hpp>
#include <boost/range/concepts.hpp>
#include <boost/range/begin.hpp>
#include <boost/range/end.hpp>
#include <ostream>

namespace bnl
{
namespace algorithm
{

/**
 * \brief Join algorithm for input iterators into output stream.
 *
 * This algorithm combines all elements inside a pair of input iterators with
 * the segments concatenated by the specified delimiter. The result is inserted
 * into the given output stream.
 *
 * \param first Input iterator to the initial position in a sequence.
 * \param last Input iterator to the final position in a sequence.
 * \param delimiter The sequence used to separate the joined segments. Various
 *        types are accepted, like a char or string.
 * \param os The output stream to place the joined result.
 * \return The same output stream.
 * \ingroup algorithm
 */
template<typename InputIterator, typename Delimiter>
::std::ostream& join(InputIterator first, InputIterator last,
                     const Delimiter& delimiter, ::std::ostream& os)
{
    BOOST_CONCEPT_ASSERT((::boost::InputIterator<InputIterator>));
    if (first != last)
    {
        os << *first;
        while (++first != last)
        {
            os << delimiter << *first;
        }
    }
    return os;
}

/**
 * \brief Join algorithm for a range into output stream.
 *
 * Boost range-based version of the join algorithm. The input range must have a
 * suitable member/free function accessible by boost::begin and boost::end.
 *
 * \pre SinglePassRange is a model of the SinglePassRangeConcept
 * \param range The range that contains elements to be joined.
 * \param delimiter The sequence used to separate the joined segments. Various
 *        types are accepted, like a char or string.
 * \param os The output stream to place the joined result.
 * \return The same output stream.
 * \see join(InputIterator, InputIterator, const Delimiter&, ::std::ostream&)
 * \ingroup algorithm
 */
template<typename SinglePassRange, typename Delimiter>
inline ::std::ostream& join(const SinglePassRange& range,
                            const Delimiter& delimiter, ::std::ostream& os)
{
    BOOST_RANGE_CONCEPT_ASSERT((::boost::SinglePassRangeConcept<const SinglePassRange>));
    return (join)(::boost::begin(range), ::boost::end(range), delimiter, os);
}

} // namespace algorithm
} // namespace bnl

#endif // BNL_ALGORITHM_JOIN_HPP
