/**
 * \file
 * \brief Defines functions to deal with Graphviz files.
 */

#ifndef BNL_GRAPHVIZ_HPP
#define BNL_GRAPHVIZ_HPP

#include <bnl/algorithm/join.hpp>
#include <bnl/bayesian_network.hpp>
#include <bnl/refcounted_dataset.hpp>
#include <bnl/score/log_score.hpp>
#include <bnl/smile_ext/idarray_range.hpp>
#include <bnl/utility/error.hpp>
#include <smile/smile.h>

#include <cstdio>
#include <cstdlib>
#include <fstream>
#include <iomanip>
#include <ios>
#include <iostream>
#include <limits>
#include <sstream>
#include <string>
#include <vector>

namespace bnl
{

/// \addtogroup graphviz
///@{

/// \cond DETAIL
namespace detail
{

template<typename LocalLogScore>
void write_graphviz_scores(
    ::std::ostream& out,
    bayesian_network& bn,
    LocalLogScore local_log_score,
    refcounted_dataset* dataset,
    bool score_order,
    unsigned max_parents,
    bool network_is_totally_ordered)
{
    if (dataset)
    {
        try
        {
            double score = log_score(bn, *dataset, local_log_score);
            out << "[Dataset]: " << dataset->name() << "\\l";
            out << "    &#x2022; [Network Log Score]: " << score << "\\l";
            if (score_order && network_is_totally_ordered)
            {
                score = log_score(bn.total_order(), *dataset,
                                  local_log_score, max_parents);
                out << "    &#x2022; [Order Log Score]:      " << score << "\\l";
            }
        }
        catch (const match_error&)
        {
            // Ignore matching error
        }
    }
}

inline void write_graphviz_scores(::std::ostream&, bayesian_network&, int,
                                  refcounted_dataset*, bool, unsigned, bool)
{
}

} // namespace detail
/// \endcond

/**
 * \brief Draw the network in some image format through Graphviz.
 *
 * The Bayesian network is first written in Graphviz dot format, and then
 * the dot command is programmatically called to generate the corresponding
 * image. The image format is taken from the input filename; however, if
 * no extension is provided, then jpg is used.
 *
 * \param filename The name of the image file that will be generated.
 * \param bn The Bayesian network to write out.
 * \param local_log_score A pointer to the local scoring function object.
 * \param dataset A pointer to the dataset to calculate a score with.
 * \param score_order Set to true in order to calculate and write the score
 *        of the internal order.
 * \param max_parents The maximum number of parents to consider.
 * \param remove_graphviz_file Whether the intermediary dot file should be
 *        deleted before returning.
 * \param precision The value to use for the floating-point precision.
 * \throw io_error if the command processor is unavailable, the intermediary
 *        file could not be opened, or any stream error.
 * \see write_graphviz
 */
template<typename LocalLogScore>
void draw_graphviz(
    ::std::string filename,
    bayesian_network& bn,
    LocalLogScore local_log_score,
    refcounted_dataset* dataset,
    bool score_order = false,
    unsigned max_parents = 5,
    bool remove_graphviz_file = true,
    int precision = ::std::numeric_limits<double>::digits10)
{
    // Make sure the command processor is available before doing anything else
    if (::std::system(0) == 0)
    {
        throw io_error("I/O Error: Command processor not available to run dot.");
    }

    // Construct temporary filename for writing Graphviz syntax
    const ::std::size_t extension_position = filename.rfind('.');
    ::std::string graphviz_filename = filename.substr(0, extension_position);
    graphviz_filename += '.';
    graphviz_filename += "gv";

    // Open temporary stream for writing dot syntax and then perform write
    {
        ::std::ofstream temp(graphviz_filename.c_str(),
                             ::std::ios::out | ::std::ios::trunc);
        if (!temp.is_open())
        {
            throw io_error("I/O Error: Unable to open intermediary file stream.");
        }
        temp << ::std::setprecision(precision);
        write_graphviz(temp, bn, local_log_score, dataset, score_order, max_parents);
        if (temp.fail())
        {
            throw io_error("I/O Error: Issue occurred while writing output stream.");
        }
    }

    // Detect the filename's extension, or choose default
    ::std::string extension;
    if (extension_position == ::std::string::npos)
    {
        extension = "jpg";
        filename += '.';
        filename += extension;
    }
    else
    {
        extension = filename.substr(extension_position + 1);
    }

    // Convert Graphviz to image and delete temporary Graphviz file
    ::std::string command("dot -T");
    command.append(extension) += ' ';
    command.append(graphviz_filename).append(" -o ").append(filename);
    int unused = ::std::system(command.c_str());
    (void) unused;
    if (remove_graphviz_file)
    {
        ::std::remove(graphviz_filename.c_str());
    }
}

/**
 * \overload
 */
inline void draw_graphviz(
    ::std::string filename,
    bayesian_network& bn,
    bool remove_graphviz_file = true,
    int precision = ::std::numeric_limits<double>::digits10)
{
    draw_graphviz<int>(filename, bn, 0, 0, false, 5, remove_graphviz_file, precision);
}

/**
 * \brief Write a Bayesian network to an output stream in Graphviz dot format.
 *
 * The Bayesian network is written in dot format, which can be fed into
 * AT&T Graphviz to draw it with an automatic layout. If a valid total order
 * exists inside the network, the order will be inserted into the stream.
 *
 * Additionally, in the case that a valid matching dataset and scoring function
 * is input, the log score will be calculated and also output.
 *
 * \param out The output stream to write the network to (e.g. file stream).
 * \param bn The Bayesian network to write out.
 * \param local_log_score A pointer to the local scoring function object.
 * \param dataset A pointer to the dataset to calculate a score with.
 * \param score_order Set to true in order to calculate and write the score
 *        of the internal order.
 * \param max_parents The maximum number of parents to consider.
 */
template<typename LocalLogScore>
void write_graphviz(
    ::std::ostream& out,
    bayesian_network& bn,
    LocalLogScore local_log_score,
    refcounted_dataset* dataset,
    bool score_order = false,
    unsigned max_parents = 5)
{
    out << "digraph G {\n";
    for (int i = bn.GetFirstNode(); i != DSL_OUT_OF_RANGE; i = bn.GetNextNode(i))
    {
        // Output all edges pointing from node to children
        const ::DSL_intArray& children = bn.GetChildren(i);
        const int num_children = children.NumItems();
        for (int j = 0; j < num_children; ++j)
        {
            out << i << " -> " << children[j] << ";\n";
        }

        // Output the identifier of the node, if available
        out << i;
        ::DSL_node* const node = bn.GetNode(i);
        if (node)
        {
            ::std::string attributes;
            const char* const id = node->Info().Header().GetId();
            if (id && id[0] != '\0')
            {
                attributes.append("label=\"").append(id) += '\"';
            }

            ::DSL_nodeDefinition* const definition = node->Definition();
            const ::DSL_idArray* outcomes_names;
            if (definition && (outcomes_names = definition->GetOutcomesNames()))
            {
                ::std::ostringstream oss;
                algorithm::join(*outcomes_names, ',', oss);
                ::std::string joined_outcomes_names = oss.str();
                if (!joined_outcomes_names.empty())
                {
                    if (!attributes.empty())
                    {
                        attributes += ',';
                    }
                    attributes.append("outcomes=\"").append(joined_outcomes_names);
                    attributes += '\"';
                }
            }

            if (!attributes.empty())
            {
                out << " [" << attributes << "]";
            }
        }
        out << ";\n";
    }

    // Store network attributes
    out << '\n';
    const char* const network_id = bn.Header().GetId();
    if (network_id && network_id[0] != '\0')
    {
        out << "name=\"" << network_id << "\";\n";
    }

    // Prepare a label for additional textual information
    out << "fontcolor=\"darkslateblue\";\n";
    out << "fontsize=11.0;\n";
    out << "labeljust=\"l\";\n";
    out << "label=\"\\l";
    if (network_id)
    {
        out << "[Network]: " << network_id << "\\l";
    }

    // Display information on total order, if valid
    const bool network_is_totally_ordered = bn.totally_ordered();
    if (network_is_totally_ordered)
    {
        typedef ::std::vector<int>::const_iterator const_iterator;
        const_iterator it = bn.total_order().begin();
        const_iterator it_end = bn.total_order().end();
        out << "[Order]: " << bn.get_id(*it);
        while (++it != it_end)
        {
            out << " &#x227a; " << bn.get_id(*it);
        }
        out << "\\l";
    }

    // Display information on scores from dataset, if valid
    detail::write_graphviz_scores(out, bn, local_log_score, dataset, score_order,
                                  max_parents, network_is_totally_ordered);

    // Close any remaining labels
    out << "\";\n}\n";
}

/**
 * \overload
 */
inline void write_graphviz(::std::ostream& out, bayesian_network& bn)
{
    write_graphviz<int>(out, bn, 0, 0);
}

/**
 * \brief Convert a GraphViz DOT graph into a Bayesian network.
 * \param in An input stream object for the DOT graph.
 * \param bn The corresponding Bayesian network.
 * \throw io_error if any issue reading DOT file or writing SMILE network.
 */
void read_graphviz(::std::istream& in, bayesian_network& bn);

/**
 * \brief Convert a GraphViz DOT graph into a Bayesian network.
 * \param in An input stream object for the DOT graph.
 * \return The corresponding Bayesian network.
 * \throw io_error if any issue reading DOT file or writing SMILE network.
 */
bayesian_network read_graphviz(::std::istream& in);

///@}

} // namespace bnl

#endif // BNL_GRAPHVIZ_HPP
