/**
 * \file
 * \brief Defines the log-add-exp function.
 */

#ifndef BNL_MATH_LOGADDEXP_HPP
#define BNL_MATH_LOGADDEXP_HPP

#include <boost/math/special_functions/log1p.hpp>
#include <boost/math/tools/promotion.hpp>
#include <cmath>

namespace bnl
{
namespace math
{

/// \cond DETAIL
namespace detail
{
template<typename T>
T logaddexp_imp(T x, T y)
{
    using ::std::exp;

    const T x_minus_y = x - y;
    if (x_minus_y > 0)
    {
        return x + (::boost::math::log1p)(exp(-x_minus_y));
    }
    else if (x_minus_y <= 0)
    {
        return y + (::boost::math::log1p)(exp(x_minus_y));
    }
    else
    {
        return x + y;
    }
}
} // namespace detail
/// \endcond

/**
 * \brief Computes natural logarithm of the sum of exp(x) and exp(y).
 *
 * The return type of this function is a floating-point type based on the input
 * types. The calculation is done in such a way to avoid undue underflow and
 * overflow.
 *
 * This implementation is based on Python numpy's function of the same name.
 *
 * \param x The first argument to exponentiate.
 * \param y The second argument to exponentiate.
 * \return \f$\mathtt{logaddexp}(x,y) = \log(e^x + e^y)\f$
 * \ingroup math
 */
template<typename T1, typename T2>
#ifdef BNL_DOXYGEN_INVOKED
calculated_return_type
#else
inline typename ::boost::math::tools::promote_args<T1, T2>::type
#endif
logaddexp(T1 x, T2 y)
{
    typedef typename ::boost::math::tools::promote_args<T1, T2>::type result_type;
    return detail::logaddexp_imp(static_cast<result_type>(x), static_cast<result_type>(y));
}

} // namespace math
} // namespace bnl

#endif // BNL_MATH_LOGADDEXP_HPP
