/**
 * \file
 * \brief Defines the log-sum-exp function.
 */

#ifndef BNL_MATH_LOGSUMEXP_HPP
#define BNL_MATH_LOGSUMEXP_HPP

#include <boost/math/special_functions/fpclassify.hpp>
#include <boost/math/tools/promotion.hpp>

#include <boost/concept_check.hpp>
#include <boost/concept/assert.hpp>

#include <boost/range/concepts.hpp>
#include <boost/range/begin.hpp>
#include <boost/range/end.hpp>
#include <boost/range/value_type.hpp>

#include <algorithm>
#include <cassert>
#include <cmath>
#include <iterator>

namespace bnl
{
namespace math
{

/**
 * \brief Computes natural logarithm of the sum of all elements exponentiated.
 *
 * The return type of this function is a floating-point type based on the input
 * types. The calculation is done in such a way to avoid undue underflow and
 * overflow.
 *
 * This implementation is a simplified version based on Python scipy's function
 * of the same name.
 *
 * \pre The input range must be non-empty.
 * \param first Forward iterator to the initial position of a sequence.
 * \param last Forward iterator to the final position of a sequence.
 * \return \f$\mathtt{logsumexp}(\mathbf{x}) = \log \sum_i e^{x_i}\f$
 * \see logaddexp
 * \ingroup math
 */
template<typename ForwardIterator>
#ifdef BNL_DOXYGEN_INVOKED
calculated_return_type
#else
typename ::boost::math::tools::promote_args<
    typename ::std::iterator_traits<ForwardIterator>::value_type>::type
#endif
logsumexp(ForwardIterator first, ForwardIterator last)
{
    BOOST_CONCEPT_ASSERT((::boost::ForwardIterator<ForwardIterator>));
    typedef typename ::std::iterator_traits<ForwardIterator>::value_type input_type;
    typedef typename ::boost::math::tools::promote_args<input_type>::type result_type;
    using ::std::exp;
    using ::std::log;

    assert(first != last);
    result_type x_max = *(::std::max_element(first, last));
    if (!(::boost::math::isfinite)(x_max))
    {
        x_max = static_cast<result_type>(0);
    }

    result_type sum = static_cast<result_type>(0);
    for (; first != last; ++first)
    {
        sum += exp(*first - x_max);
    }

    return x_max + log(sum);
}

/**
 * \brief Computes the log-sum-exp of a Boost.Range.
 * \pre The input range must be non-empty.
 * \param range The range containing elements to run the log-sum-exp on.
 * \return \f$\mathtt{logsumexp}(\mathbf{x}) = \log \sum_i e^{x_i}\f$
 * \see logsumexp(ForwardIterator, ForwardIterator)
 * \ingroup math
 */
template<typename ForwardRange>
#ifdef BNL_DOXYGEN_INVOKED
calculated_return_type
#else
inline typename ::boost::math::tools::promote_args<
    typename ::boost::range_value<ForwardRange>::type>::type
#endif
logsumexp(ForwardRange& range)
{
    BOOST_RANGE_CONCEPT_ASSERT((::boost::ForwardRangeConcept<ForwardRange>));
    return logsumexp(::boost::begin(range), ::boost::end(range));
}

/**
 * \overload
 * \ingroup math
 */
template<typename ForwardRange>
#ifdef BNL_DOXYGEN_INVOKED
calculated_return_type
#else
inline typename ::boost::math::tools::promote_args<
    typename ::boost::range_value<const ForwardRange>::type>::type
#endif
logsumexp(const ForwardRange& range)
{
    BOOST_RANGE_CONCEPT_ASSERT((::boost::ForwardRangeConcept<const ForwardRange>));
    return logsumexp(::boost::begin(range), ::boost::end(range));
}

} // namespace math
} // namespace bnl

#endif // BNL_MATH_LOGSUMEXP_HPP
