/**
 * \file
 * \ingroup random
 * \brief Defines the random rotate algorithm using global RNG.
 */

#ifndef BNL_RANDOM_ALGORITHM_RANDOM_ROTATE_HPP
#define BNL_RANDOM_ALGORITHM_RANDOM_ROTATE_HPP

#include <bnl/random/randgen.hpp>

#include <boost/concept/assert.hpp>
#include <boost/concept_check.hpp>

#include <boost/range/concepts.hpp>
#include <boost/range/begin.hpp>
#include <boost/range/end.hpp>
#include <boost/range/difference_type.hpp>

#include <algorithm>
#include <iterator>

namespace bnl
{
namespace algorithm
{

/// \addtogroup algorithm
///@{

/**
 * \brief Rotates the elements in the range [first,last) by a random amount.
 *
 * All possible unique states from rotating should be equally likely. If the
 * initial state (i.e. a zero rotation) should be ignored, the final parameter
 * should be set to true.
 *
 * \note In the case that there are insufficient elements to generate a
 *       non-initial state, the final parameter will be ignored.
 * \param first Forward iterator to the initial position of a sequence.
 * \param last Forward iterator to the final position of the sequence.
 * \param ignore_zero_rotate Whether to ignore the possibility of zero rotation.
 * \return The amount rotated.
 */
template<typename ForwardIterator>
typename ::std::iterator_traits<ForwardIterator>::difference_type
random_rotate(ForwardIterator first, ForwardIterator last, bool ignore_zero_rotate = true)
{
    BOOST_CONCEPT_ASSERT((::boost::ForwardIterator<ForwardIterator>));
    typedef typename ::std::iterator_traits<ForwardIterator>::difference_type diff_type;
    const diff_type size = ::std::distance(first, last);
    if (size > 1)
    {
        const int x = static_cast<int>(ignore_zero_rotate);
        const diff_type amount = static_cast<diff_type>(random::global_rng()(size - x) + x);
        ForwardIterator middle = first;
        ::std::advance(middle, amount);
        ::std::rotate(first, middle, last);
        return amount;
    }
    else
    {
        return 0;
    }
}

/**
 * \brief Random rotate algorithm for Boost.Range.
 * \pre ForwardRange is a model of the ForwardRangeConcept.
 * \param range The range to perform the operation on.
 * \param ignore_zero_rotate Whether to ignore the possibility of zero rotation.
 * \return The amount rotated.
 * \see random_rotate(ForwardRange, ForwardRange, bool)
 */
template<typename ForwardRange>
inline typename ::boost::range_difference<ForwardRange>::type
random_rotate(ForwardRange& range, bool ignore_zero_rotate = true)
{
    BOOST_RANGE_CONCEPT_ASSERT((::boost::ForwardRangeConcept<ForwardRange>));
    return (random_rotate)(::boost::begin(range), ::boost::end(range), ignore_zero_rotate);
}

/**
 * \overload
 */
template<typename ForwardRange>
inline typename ::boost::range_difference<const ForwardRange>::type
random_rotate(const ForwardRange& range, bool ignore_zero_rotate = true)
{
    BOOST_RANGE_CONCEPT_ASSERT((::boost::ForwardRangeConcept<const ForwardRange>));
    return (random_rotate)(::boost::begin(range), ::boost::end(range), ignore_zero_rotate);
}

///@}

} // namespace algorithm
} // namespace bnl

#endif // BNL_RANDOM_ALGORITHM_RANDOM_ROTATE_HPP
