/**
 * \file
 * \ingroup random
 * \brief Defines the random shuffle algorithm to use a global RNG.
 */

#ifndef BNL_RANDOM_ALGORITHM_RANDOM_SHUFFLE_HPP
#define BNL_RANDOM_ALGORITHM_RANDOM_SHUFFLE_HPP

#include <bnl/random/randgen.hpp>
#include <boost/concept/assert.hpp>
#include <boost/concept_check.hpp>
#include <boost/range/algorithm/random_shuffle.hpp>
#include <algorithm>

namespace bnl
{
namespace algorithm
{

/// \addtogroup algorithm
///@{

/**
 * \brief Rearranges the elements in the range [first,last) randomly.
 *
 * This makes use of the global RNG rather than the implementation-specific,
 * unspecified version.
 *
 * \param first Forward iterator to the initial position of a sequence.
 * \param last Forward iterators to the final position of the sequence.
 * \see std::random_shuffle
 */
template<typename RandomAccessIterator>
inline void random_shuffle(RandomAccessIterator first, RandomAccessIterator last)
{
    BOOST_CONCEPT_ASSERT((::boost::RandomAccessIterator<RandomAccessIterator>));
    ::std::random_shuffle(first, last, random::global_rng());
}

/**
 * \brief Random shuffle algorithm for Boost.Range.
 *
 * Boost range-based version of the shuffle implementation using a global RNG.
 *
 * \pre RandomAccessRange is a model of the RandomAccessRangeConcept.
 * \param range The range to be shuffled.
 * \return The same range after shuffling.
 */
template<typename RandomAccessRange>
inline RandomAccessRange& random_shuffle(RandomAccessRange& range)
{
    return ::boost::range::random_shuffle(range, random::global_rng());
}

/**
 * \overload
 */
template<typename RandomAccessRange>
inline const RandomAccessRange& random_shuffle(const RandomAccessRange& range)
{
    return ::boost::range::random_shuffle(range, random::global_rng());
}

///@}

} // namespace algorithm
} // namespace bnl

#endif // BNL_RANDOM_ALGORITHM_RANDOM_SHUFFLE_HPP
