/**
 * \file
 * \brief Defines various random number generators.
 *
 * Random number generation is important in many learning algorithms.
 * This file tries to guarantee a uniform distribution for any interval with
 * reasonable quality randomness, compared to C/C++'s std::rand() function.
 *
 * The Boost.Random library is used to generate random numbers for usage in
 * pre-C++11/0x code.
 */

#ifndef BNL_RANDOM_RANDGEN_HPP
#define BNL_RANDOM_RANDGEN_HPP

#include <boost/random/mersenne_twister.hpp>
#include <boost/random/random_number_generator.hpp>
#include <boost/version.hpp>

#if BOOST_VERSION >= 104700
#include <boost/random/uniform_int_distribution.hpp>
#include <boost/random/uniform_real_distribution.hpp>
#else
#include <boost/random/uniform_int.hpp>
#include <boost/random/uniform_real.hpp>
#endif

#include <cstddef>

namespace bnl
{
namespace random
{

/// \addtogroup random
///@{

/// \cond
#if BOOST_VERSION >= 104700
typedef ::boost::random::uniform_int_distribution<int> uniform_int_distribution;
typedef ::boost::random::uniform_real_distribution<double> uniform_real_distribution;
typedef ::boost::random::mt19937 urng_type;
typedef ::boost::random::random_number_generator<urng_type, ::std::ptrdiff_t> rng_type;
#else
typedef ::boost::uniform_int<int> uniform_int_distribution;
typedef ::boost::uniform_real<double> uniform_real_distribution;
typedef ::boost::mt19937 urng_type;
typedef ::boost::random_number_generator<urng_type, ::std::ptrdiff_t> rng_type;
#endif
/// \endcond

/// \name Global generators
///@{
/**
 * \brief Access a global instantiation of UniformRandomNumberGenerator concept.
 *
 * The URNG is lazily constructed upon the first call to the function. The
 * function is not designed in a thread-safe manner, so care may be needed
 * in a multi-threaded environment.
 *
 * This implementation is based on the open standard proposal WG21 N3847.
 *
 * \return A global URNG object.
 */
urng_type& global_urng();

/**
 * \brief Access a global instantiation of RandomNumberGenerator concept.
 *
 * The corresponding concept is from std:25.2.11 [lib.alg.random.shuffle].
 * The resulting RNG object can intake and output a std::ptrdiff_t for use with
 * iterators.
 *
 * \return A global RNG object.
 */
rng_type& global_rng();
///@}

/// \name Ranged generators
///@{
/**
 * \brief A uniform random-number generator for real values.
 * \pre min_arg <= max_arg
 * \param min_arg The minimum bound of the interval to generate a random value.
 * \param max_arg The maximum bound of the interval to generate a random value.
 * \return A random real value in the range [min_arg, max_arg).
 */
double random_real(double min_arg = 0.0, double max_arg = 1.0);

/**
 * \brief A uniform random-number generator for integer values.
 * \pre min_arg <= max_arg
 * \param min_arg The minimum bound of the interval to generate a random value.
 * \param max_arg The maximum bound of the interval to generate a random value.
 * \return A random int value in the range [min_arg, max_arg].
 */
int random_int(int min_arg = 0, int max_arg = 1);

/**
 * \brief A uniform random-number generator for boolean values.
 * \return A random boolean value (i.e. true or false).
 */
bool random_bool();
///@}

///@}

} // namespace random
} // namespace bnl

#endif // BNL_RANDOM_RANDGEN_HPP
