/**
 * \file
 * \brief Defines the base reference-counted dataset class.
 *
 * Functionality is provided for dealing with datasets in reference-counted
 * format. Algorithms are implemented to score the dataset given a variety of
 * possible inputs.
 */

#ifndef BNL_REFCOUNTED_DATASET_HPP
#define BNL_REFCOUNTED_DATASET_HPP

#include <smile/smilearn.h>
#include <string>
#include <vector>

class DSL_network;

namespace bnl
{

/**
 * \brief Reference-counted dataset where variable states are nonnegative ints.
 * \ingroup dataset
 *
 * The states for a variable start from zero and following the successor
 * relation. So, the largest observed state provides sufficient information
 * to detect the total number of possible states.
 *
 * This base class provides simple query abilities that minimize memory usage.
 * Accelerated versions can be implemented as derived classes that implement
 * all virtual functions.
 */
class refcounted_dataset
{
public:
/// \name (constructor)
//@{
    /**
     * \brief Constructs an empty reference-counted dataset.
     *
     * Simply sets up the reference-counted dataset object with no initial data.
     * A dataset can be loaded later on by using the read function.
     */
    refcounted_dataset();

    /**
     * \brief Constructs a reference-counted dataset from a SMILE dataset.
     *
     * Sets up the reference-counted dataset object with the same data as the
     * given SMILE DSL_dataset object. Will throw an exception on error.
     *
     * \param dataset The existing SMILE dataset object.
     * \param name A name to attach to this dataset.
     * \param remove_constants Optional argument to remove constant random
     *        variables. It is false by default.
     * \throw io_error if the dataset is not discrete.
     */
    refcounted_dataset(const ::DSL_dataset& dataset,
                       const ::std::string& name = "",
                       bool remove_constants = false);

    /**
     * \brief Constructs a reference-counted dataset from a data file.
     *
     * Sets up the reference-counted dataset object with the same data as given
     * inside the specified file. Will throw an exception on error.
     *
     * \param filename Name of file containing data in SMILE-readable format.
     * \param remove_constants Optional argument to remove constant random
     *        variables. It is false by default.
     * \throw io_error if unable to read file or if the dataset is not discrete.
     * \see read
     */
    explicit refcounted_dataset(const ::std::string& filename,
                                bool remove_constants = false);
//@}

/// \name (destructor)
//@{
    virtual ~refcounted_dataset();
//@}

/// \name Modifiers
//@{
    /**
     * \brief Read in a dataset from the specified file.
     *
     * The data file corresponding to the input filename is parsed by SMILE and
     * loaded into memory. The data must contain only discrete variables;
     * otherwise, an exception will be thrown.
     *
     * \param filename Name of file containing data in SMILE-readable format.
     * \param remove_constants Set in order to remove constant random variables.
     * \throw io_error if unable to read file or if the dataset is not discrete.
     */
    virtual void read(const ::std::string& filename, bool remove_constants);

    /**
     * \brief Remove constant random variables from dataset.
     *
     * Searches the underlying dataset for any single-state (i.e. constant)
     * random variables and then removes them. This is used to eliminate
     * unnecessary variables that provide no additional information. Ignores
     * any SMILE errors.
     *
     * \return Number of variables removed from dataset.
     */
    virtual unsigned remove_constant_variables();
//@}

/// \name Queries
//@{
    /**
     * \brief Get the name associated with this dataset.
     * \return The name based on the filename the data was loaded from.
     */
    const ::std::string& name() const;

    /**
     * \brief Get the number of states for a particular random variable.
     *
     * Returns the number of states for the given random variable based on
     * assumptions made for reference-counted data. No bounds-checking is done,
     * so a portable program should never call this function with an argument
     * that is out of range.
     *
     * \param var The random variable that is being queried.
     * \return The number of states for a particular random variable.
     */
    unsigned num_states(int var) const;

    /**
     * \brief Count the number of entries that satisfy the constraints (n_ijk).
     *
     * The constraints are given as a configuration that applies to a subset of
     * all variables. So, for each index i in the given inputs, it must be the
     * case that (variables[i] -> config[i]) inside the entry to be considered
     * part of the count.
     *
     * It should be the case that the size of both inputs are the same.
     *
     * \pre The number of variables must match the size of the configuration.
     * \param variables The subset of all variables that has a constraint.
     * \param config The configuration the variables must be in.
     * \return The number of entries in the dataset that match the constraints.
     */
    virtual unsigned count(const ::std::vector<int>& variables,
                           const ::std::vector<unsigned>& config) const;

    /**
     * \brief Access the underlying SMILE dataset.
     * \return The underlying dataset.
     */
    const ::DSL_dataset& data() const;
//@}

/// \name Network-Related
//@{
    /**
     * \brief Set up nodes in the Bayesian network to match the dataset.
     *
     * SMILE requires all node states to start with a letter, so the default
     * prefix "S" is used. The underlying DSL_dataset may be modified as
     * mentioned by the SMILE API.
     *
     * \param bn The Bayesian network to set up the structure inside.
     */
    void prepare_structure(::DSL_network& bn);

    /**
     * \brief Finds all matches between the dataset variables and network nodes.
     * \param bn The Bayesian network to detect matches with.
     * \return All matches between the dataset and network.
     * \throw match_error if the network and dataset fail to match.
     */
    ::std::vector< ::DSL_datasetMatch> match_network(const ::DSL_network& bn);
//@}

/// \name Order-Related
//@{
    /**
     * \brief Generate a random total order from the random variables.
     *
     * A random total order is constructed containing all the random variables
     * in the dataset (indexed from 0 to n-1). The order is stored in a vector
     * where every adjacent pair of values follows the "is not child of"
     * relation---e.g. (A,B) is a pair in the vector [...,A,B,...] and
     * represents that A is not a child of B.
     *
     * \return The generated random total order as a vector.
     */
    ::std::vector<int> random_order() const;

    /**
     * \brief Check if a given total order can be used on this dataset.
     *
     * The input total order is tested to make sure that all variables inside
     * the order are also inside the dataset and that no variable occurs more
     * than once.
     *
     * \param order The total order to test with.
     * \return true if the total order can be used; otherwise, false.
     */
    bool can_use_order(const ::std::vector<int>& order) const;
//@}

protected:
    /**
     * \brief The underlying SMILE dataset object.
     *
     * The SMILE implementation already had the necessary storage capabilities
     * and ability to parse user input files, so it is added internally.
     */
    ::DSL_dataset dataset_;

private:
    void init(bool remove_constants);
    ::std::string name_;
};

} // namespace bnl

#endif // BNL_REFCOUNTED_DATASET_HPP
