/**
 * \file
 * \brief Defines the log BDeu scoring equations for local nodes/variables.
 */

#ifndef BNL_SCORE_LOCAL_LOG_BDEU_SCORE
#define BNL_SCORE_LOCAL_LOG_BDEU_SCORE

#include <bnl/treecount.hpp>
#include <vector>

namespace bnl
{

/**
 * \brief A functor for computing the local log BDeu score.
 * \ingroup score
 *
 * This provides functionality to compute the log score for a particular
 * local region (one node and its parents). The computation uses the assumptions
 * described in the BDeu scoring function.
 */
class local_log_bdeu_score
{
public:
    /**
     * \brief Construct a functor to calculate log BDeu score of a variable.
     * \param ess The equivalent sample size.
     */
    local_log_bdeu_score(double ess = 1.0);

    /**
     * \brief Compute the log BDeu score for a node/variable.
     *
     * The variable to score is passed in as the zero-th index of all variables,
     * while all parents are passed in the other indicies. For example, [3,1,5]
     * means that variable 3 has variables 1 and 5 as parents.
     *
     * The log score is computed using the BDeu metric's priors given by the
     * equation \f$\alpha_{ijk} = \frac{\alpha^*}{q_i r_i}\f$, where
     * \f$\alpha^*\f$ is the equivalent sample size (ess).
     *
     * \pre There must be at least one element inside the input variables.
     * \param dataset A reference to an existing reference-counting dataset.
     * \param variables variables[0] is the node while variables[1..(n-1)] are
     *        the parents of the node.
     * \return The log BDeu score of the node.
     */
    double operator()(const TreeCount& root,
                      const ::std::vector<int>& variables, int NumOfVars) const;

private:
    double ess_;
};

} // namespace bnl

#endif // BNL_SCORE_LOCAL_LOG_BDEU_SCORE
