/**
 * \file
 * \brief Defines the log K2 scoring equations for local nodes/variables.
 */

#ifndef BNL_SCORE_LOCAL_LOG_K2_SCORE
#define BNL_SCORE_LOCAL_LOG_K2_SCORE

#include <bnl/refcounted_dataset.hpp>
#include <vector>

namespace bnl
{

/**
 * \brief A functor for computing the local log K2 score.
 * \ingroup score
 *
 * This provides functionality to compute the log score for a particular
 * local region (one node and its parents). The computation uses the assumptions
 * described in the K2 scoring function.
 */
class local_log_k2_score
{
public:
    /**
     * \brief Construct a functor to calculate log K2 score of a variable.
     */
    local_log_k2_score();

    /**
     * \brief Compute the log K2 score for a node/variable.
     *
     * The variable to score is passed in as the zero-th index of all variables,
     * while all parents are passed in the other indicies. For example, [3,1,5]
     * means that variable 3 has variables 1 and 5 as parents.
     *
     * The log score is computed using the K2 metric's constant prior,
     * i.e. \f$\alpha_{ijk} = 1\f$.
     *
     * \pre There must be at least one element inside the input variables.
     * \param dataset A reference to an existing reference-counting dataset.
     * \param variables variables[0] is the node while variables[1..(n-1)] are
     *        the parents of the node.
     * \return The log K2 score of the node.
     */
    double operator()(const refcounted_dataset& dataset,
                      const ::std::vector<int>& variables) const;
};

} // namespace bnl

#endif // BNL_SCORE_LOCAL_LOG_K2_SCORE
