/**
 * \file
 * \brief Defines general log scoring equations.
 */

#ifndef BNL_SCORE_LOG_SCORE
#define BNL_SCORE_LOG_SCORE

#include <bnl/math/logaddexp.hpp>
#include <bnl/refcounted_dataset.hpp>
#include <bnl/smile_ext/dataset_match.hpp>
#include <bnl/utility/error.hpp>
#include <bnl/utility/odometer.hpp>
#include <bnl/treecount.hpp>

#include <smile/smile.h>
#include <smile/smilearn.h>
#include <iterator>
#include <vector>

namespace bnl
{

/// \addtogroup score
///@{

/**
 * \brief Compute the log score of a Bayesian network.
 * \param bn The Bayesian network to score.
 * \param dataset The reference-counted dataset used for scoring.
 * \param local_log_score The function object for computing local score.
 * \return The score of the network.
 * \throws match_error if the dataset and network do not match.
 */
template<typename LocalLogScore>
double log_score(const ::DSL_network& bn, refcounted_dataset& dataset, 
                 LocalLogScore local_log_score)
{
    typedef ::std::vector< ::DSL_datasetMatch>::const_iterator match_iterator;
    const ::std::vector< ::DSL_datasetMatch> matches = dataset.match_network(bn);
    const ::std::vector<int> node_to_variable = smile::make_vector_map(matches);

    double score = 0.0;
    ::std::vector<int> variables;
    for (match_iterator it = matches.begin(); it != matches.end(); ++it)
    {
        const ::DSL_intArray& parents = bn.GetParents(it->node);
        const int num_parents = parents.NumItems();
        variables.resize(num_parents + 1);
        variables[0] = it->column;
        for (int i = 0; i < num_parents; ++i)
        {
            variables[i + 1] = node_to_variable[parents[i]];
        }

        score += local_log_score(dataset, variables);
    }
    return score;
}

/**
 * \brief Compute the log score of a total order.
 * \param order The total order to score.
 * \param dataset The reference-counted dataset used for scoring.
 * \param local_log_score The function object for computing local (node) score.
 * \param max_parents The maximum number of parents to consider.
 * \return The score of the total order.
 * \throws match_error if the dataset and order do not match.
 */
template<typename LocalLogScore>
double log_score(const ::std::vector<int>& order, const refcounted_dataset& dataset, const TreeCount& root,
                 LocalLogScore local_log_score, int NumOfVars, unsigned max_parents = 3)
{
    if (!dataset.can_use_order(order))
    {
        throw match_error("Match Error: The total order and dataset do not match");
    }

    // Initialize variables for keeping track
    typedef ::std::vector<int>::const_iterator const_iterator;
    const const_iterator first = order.begin();
    const const_iterator last = order.end();
    ::std::vector<int> variables(1);
    double score = 0.0;
    odometer<> enabled;
    odometer<>::size_type num_possible_parents = 0;
	
    // Score the order variable by variable
    for (const_iterator current = first; current != last; ++current)
    {
		//std::cout << "Running..." << '\n';
        enabled.clear_rollover();
        enabled.resize(num_possible_parents, 1, 0);
        variables.resize(1);
        variables[0] = *current;
        double s = local_log_score(root, variables, NumOfVars);
        //std::cout << "Local log score:    " << s << '\n';
        while (!(++enabled).rollover())
        {
            const unsigned num_parents = enabled.sum_digits<unsigned>();
            
            if (num_parents <= max_parents)
            {
                variables.resize(num_parents + 1);
                unsigned j = 0;
                const_iterator parent = first;
                for (odometer<>::size_type k = 0; j < num_parents; ++k, ++parent)
                {
                    if (enabled[k] == 1)
                    {
                        variables[++j] = *parent;
                    }
                }
                
                //std::cout << "Local log score 3:    " << variables[0] << '\t' << variables[1] << '\n';
                s = math::logaddexp(s, local_log_score(root, variables, NumOfVars));
               // std::cout << "Local log score 2:    " << s << '\n';
            }
        }
        score += s;
        //std::cout << "Score:    " << s << '\n';
        ++num_possible_parents;
    }
    return score;
}



///@}

} // namespace bnl

#endif // BNL_SCORE_LOG_SCORE
