/**
 * \file
 * \brief Defines an odometer to cycle through all states.
 */

#ifndef BNL_UTILITY_ODOMETER_HPP
#define BNL_UTILITY_ODOMETER_HPP

#include <bnl/algorithm/join.hpp>

#include <boost/concept_check.hpp>
#include <boost/concept/assert.hpp>
#include <boost/iterator/zip_iterator.hpp>
#include <boost/mpl/assert.hpp>
#include <boost/tuple/tuple.hpp>
#include <boost/type_traits.hpp>
#include <boost/utility/enable_if.hpp>

#include <algorithm>
#include <cassert>
#include <functional>
#include <numeric>
#include <ostream>
#include <vector>

namespace bnl
{

/// \addtogroup utility
///@{

/// \cond DETAIL
namespace detail
{
template<typename T>
T succ(T x = T())
{
    BOOST_MPL_ASSERT((::boost::has_pre_increment<T>));
    return ++x;
}
} // namespace detail
/// \endcond

/**
 * \brief The odometer class imitates the behavior of a car's odometer.
 *
 * This class is designed in a generalized manner for many types. Additionally,
 * each digit can be set to a different value (unlike pure decimal values for a
 * car). All possible values on the odometer can be seen by incrementing.
 *
 * Example:
 * If the maximum is set to [9,9,9] and the state starts at [0,0,0], then the
 * next value should be [1,0,0] (given the left-most value is the zero-index of
 * the array) and the last value should be [9,9,9]. When another increment is
 * performed, the value wraps around to [0,0,0] and the rollover flag is set.
 *
 * \pre T must meet the requirements of the concepts DefaultConstructible,
 *      Assignable, EqualityComparable, and Comparable (SGI LessThanComparable).
 */
template<typename T = unsigned>
class odometer
{
    BOOST_CONCEPT_ASSERT((::boost::DefaultConstructible<T>));
    BOOST_CONCEPT_ASSERT((::boost::Assignable<T>));
    BOOST_CONCEPT_ASSERT((::boost::EqualityComparable<T>));
    BOOST_CONCEPT_ASSERT((::boost::Comparable<T>));

public:
//@{
    typedef T value_type;
    typedef ::boost::zip_iterator< ::boost::tuple<
        typename ::std::vector<T>::iterator,
        typename ::std::vector<T>::iterator > > iterator;
    typedef ::boost::zip_iterator< ::boost::tuple<
        typename ::std::vector<T>::const_iterator,
        typename ::std::vector<T>::const_iterator > > const_iterator;
    typedef T& reference;
    typedef const T& const_reference;
    typedef typename ::std::vector<T>::size_type size_type;
    typedef typename ::std::vector<T>::difference_type difference_type;
//@}

/// \name (constructor)
//@{
    /**
     * \brief Empty odometer constructor (default constructor).
     *
     * Constructs an odometer with no digits.
     */
    odometer();

    /**
     * \brief Fill constructor.
     *
     * Constructs an odometer with n digits, where each maximum is set to max
     * and each current state is set to curr.
     *
     * \param n The number of digits to initialize the odometer with.
     * \param max The maximum value set for each digit. In order to use the
     *        default value, there must be a constructor that can take integers.
     * \param curr The current state value set for each digit.
     */
    explicit odometer(size_type n, const T& max = detail::succ<T>(),
                      const T& curr = T());

    /**
     * \brief Range constructor.
     *
     * Constructs an odometer with as many digits as the range [first,last),
     * with each maximum digit set to the corresponding value in that range.
     *
     * \param first Input iterators to the initial position inside a range.
     * \param last Input iterators to the final position inside a range.
     */
    template<typename InputIterator>
    odometer(InputIterator first, InputIterator last
             /// \cond
             , typename ::boost::disable_if<
                   ::boost::is_arithmetic<InputIterator> >::type* = 0
             /// \endcond
            );
//@}

/// \name Capacity
//@{
    /**
     * \brief Resize the odometer so that it contains exactly n digits.
     *
     * If n is smaller than the current size, then only the first n digits are
     * retained while all others are removed.
     *
     * If n is greater than the current size, then more digits are added with
     * the maximum initialized to the value of max and the current state
     * initialized to the value of curr.
     *
     * It is assumed that (max >= curr). No guarantee is offered for correct
     * behavior if this invariant is violated.
     *
     * \param n The new number of digits in the odometer.
     * \param max The maximum value set for each digit. In order to use the
     *        default value, there must be a constructor that can take integers.
     * \param curr The current state value set for each new digit.
     */
    void resize(size_type n, const T& max = detail::succ<T>(), const T& curr = T());

    /**
     * \brief Return the number of digits in the odometer.
     * \return The number of digits in this odometer.
     */
    size_type size() const;
//@}

/// \name Getters & Queries
//@{
    /**
     * \brief Check if an odometer rollover occurred.
     *
     * This class defines an odometer rollover as when an increment takes the
     * state from the maximum state to the zero state, or when a decrement takes
     * the state from the zero state to the maximum state.
     *
     * \return Whether a rollover has occurred since construction or last reset.
     */
    bool rollover() const;

    /**
     * \brief Check if the odometer's current state is valid.
     *
     * In order for the current state to be valid, each digit's current value
     * must be at most the digit's maximum value.
     *
     * \return true if the current state is valid; otherwise, false.
     */
    bool valid() const;

    /**
     * \brief Get the current state of the odometer.
     * \return The current state of the odometer.
     */
    const ::std::vector<T>& current() const;

    /**
     * \brief Get the maximum state of the odometer.
     * \return The maximum state of the odometer.
     */
    const ::std::vector<T>& maximum() const;

    /**
     * \brief Computes the sum of the digits inside the odometer.
     *
     * This can only be used with arithmetic odometers and return type.
     * In other cases, a compile-time error should occur.
     *
     * Example usage:
     * \code
     * long sum = odo.sum_digits<long>();
     * float sumf = odo.sum_digits<float>();
     * \endcode
     *
     * \return The sum of the digits in the explicitly provided type.
     */
    template<typename ReturnType>
    ReturnType sum_digits() const;

    /**
     * \brief Computes the product of the digits inside the odometer.
     *
     * This can only be used with arithmetic odometers and return type.
     * In other cases, a compile-time error should occur.
     *
     * Example usage:
     * \code
     * long product = odo.product_digits<long>();
     * float productf = odo.product_digits<float>();
     * \endcode
     *
     * \return The product of the digits in the explicitly provided type.
     */
    template<typename ReturnType>
    ReturnType product_digits() const;
//@}

/// \name Digit access
//@{
    /**
     * \brief Access the maximum state of a digit.
     *
     * This should only be called with an argument n that is in the range
     * [0, size); otherwise, it will cause undefined behavior.
     *
     * \param n The particular digit to access.
     * \return A constant reference to the maximum value of a digit.
     */
    const_reference maximum(size_type n) const;

    /**
     * \brief Access the maximum state of a digit.
     *
     * This should only be called with an argument n that is in the range
     * [0, size); otherwise, it will cause undefined behavior.
     *
     * If a new maximum is set, it may be necessary to alter the current state
     * value to have a valid odometer configuration. This can be done with
     * either the array subscript operator or the reset function.
     *
     * \param n The particular digit to access.
     * \return A reference to the maximum value of a digit.
     * \see operator[](size_type), reset
     */
    reference maximum(size_type n);

    /**
     * \brief Access the current state of a particular digit.
     *
     * This should only be called with an argument n that is in the range
     * [0, size); otherwise, it will cause undefined behavior.
     *
     * \param n The particular digit to access.
     * \return A constant reference of the current state of a digit.
     */
    const_reference operator[](size_type n) const;

    /**
     * \brief Access the current state of a particular digit.
     *
     * This should only be called with an argument n that is in the range
     * [0, size); otherwise, it will cause undefined behavior.
     *
     * \param n The particular digit to access.
     * \return A reference to the current state of a digit.
     */
    reference operator[](size_type n);
//@}

/// \name Iterators
//@{
    /**
     * \brief Returns an iterator to the first element of the odometer.
     *
     * An element of the iterator is a 2-tuple where the first part is for the
     * current state and the second part is for the maximum state.
     *
     * \return Iterator to the first element.
     */
    iterator begin();

    /**
     * \overload
     */
    const_iterator begin() const;

    /**
     * \brief Returns a constant iterator to the first element of the odometer
     * \return Constant iterator to the first element
     */
    const_iterator cbegin() const;

    /**
     * \brief Returns an iterator to the element following the last element.
     *
     * An element of the iterator is a 2-tuple where the first part is for the
     * current state and the second part is for the maximum state.
     *
     * \return Iterator to the element following the last element.
     */
    iterator end();

    /**
     * \overload
     */
    const_iterator end() const;

    /**
     * \brief Returns a constant iterator to the element following the last.
     * \return Constant iterator to the element following the last element.
     */
    const_iterator cend() const;
//@}

/// \name State update
//@{
    /**
     * \brief Set odometer's state back to zero and clear rollover flag.
     */
    void reset();

    /**
     * \brief Set odometer's state to the maximum and clear rollover flag.
     */
    void reset_to_max();

    /**
     * \brief Clear the odometer rollover flag.
     */
    void clear_rollover();

    /**
     * \brief Increment the odometer.
     *
     * The value of the odometer should be set to one more than the current
     * value (or back to the zero state upon rollover). The new odometer state
     * is returned.
     *
     * \return The new odometer state.
     */
    odometer<T>& operator++();

    /**
     * \brief Increment the odometer.
     *
     * The value of the odometer should be set to one more than the current
     * value (or back to the zero state upon rollover). The old odometer state
     * is returned. For performance purposes, you should typically use the
     * prefix increment operator.
     *
     * \return The old odometer state.
     * \see operator++()
     */
    odometer<T> operator++(int);

    /**
     * \brief Decrement the odometer.
     *
     * The value of the odometer should be set to one less than the current
     * value (or to the maximum state upon rollover). The new odometer state
     * is returned.
     *
     * \return The new odometer state.
     */
    odometer<T>& operator--();

    /**
     * \brief Decrement the odometer.
     *
     * The value of the odometer should be set to one less than the current
     * value (or to the maximum state upon rollover). The old odometer state
     * is returned. For performance purposes, you should typically use the
     * prefix decrement operator.
     *
     * \return The old odometer state.
     * \see operator--()
     */
    odometer<T> operator--(int);
//@}

private:
    ::std::vector<T> maximum_;
    ::std::vector<T> current_;
    bool rollover_;
};

/**
 * \brief Insert the odometer data into a stream.
 * \param os ostream object where to insert odometer information.
 * \param obj odometer object that is to be inserted into the stream.
 * \return The same as parameter os.
 */
template<typename T>
::std::ostream& operator<<(::std::ostream& os, const odometer<T>& obj)
{
    BOOST_MPL_ASSERT((::boost::has_left_shift< ::std::ostream, T>));
    os << "Current: [";
    algorithm::join(obj.current(), ',', os);
    os << "],  Maximum: [";
    algorithm::join(obj.maximum(), ',', os);
    os << "],  Rollover: " << (obj.rollover() ? "true" : "false");
    return os;
}

/// \name Comparison operators (odometer)
///@{
/**
 * \brief Check equality of two odometers.
 *
 * This is performed by making sure the maximum states, current states,
 * and rollover flags are equal.
 *
 * \param lhs odometer to the left-hand side of the operator.
 * \param rhs odometer to the right-hand side of the operator.
 * \return true if the odometers are equal, and false otherwise.
 */
template<typename T>
bool operator==(const odometer<T>& lhs, const odometer<T>& rhs)
{
    return (lhs.rollover() == rhs.rollover() &&
            lhs.current()  == rhs.current()  &&
            lhs.maximum()  == rhs.maximum());
}

template<typename T>
bool operator!=(const odometer<T>& lhs, const odometer<T>& rhs)
{
    return !(lhs == rhs);
}
///@}

///@}
//------------------------------------------------------------------------------
/// \cond

template<typename T>
odometer<T>::odometer()
    : rollover_(false)
{
}

template<typename T>
odometer<T>::odometer(size_type n, const T& max, const T& curr)
    : maximum_(n, max), current_(n, curr), rollover_(false)
{
    assert(max >= curr);
}

template<typename T>
template<typename InputIterator>
odometer<T>::odometer(InputIterator first, InputIterator last,
    typename ::boost::disable_if< ::boost::is_arithmetic<InputIterator> >::type*)
    : maximum_(first, last), current_(maximum_.size()), rollover_(false)
{
    BOOST_CONCEPT_ASSERT((::boost::InputIterator<InputIterator>));
    assert(maximum_.size() == current_.size());
}

template<typename T>
void odometer<T>::resize(size_type n, const T& max, const T& curr)
{
    assert(max >= curr);
    maximum_.resize(n, max);
    current_.resize(n, curr);
}

template<typename T>
typename odometer<T>::size_type odometer<T>::size() const
{
    assert(maximum_.size() == current_.size());
    return maximum_.size();
}

template<typename T>
bool odometer<T>::rollover() const
{
    return rollover_;
}

template<typename T>
bool odometer<T>::valid() const
{
    assert(maximum_.size() == current_.size());
    for (size_type i = 0; i < maximum_.size(); ++i)
    {
        if (current_[i] > maximum_[i])
        {
            return false;
        }
    }
    return true;
}

template<typename T>
const ::std::vector<T>& odometer<T>::current() const
{
    return current_;
}

template<typename T>
const ::std::vector<T>& odometer<T>::maximum() const
{
    return maximum_;
}

template<typename T>
template<typename ReturnType>
ReturnType odometer<T>::sum_digits() const
{
    BOOST_MPL_ASSERT((::boost::is_convertible<int, ReturnType>));
    BOOST_MPL_ASSERT((::boost::has_plus<ReturnType, T>));
    return ::std::accumulate(current_.begin(), current_.end(),
                             static_cast<ReturnType>(0));
}

template<typename T>
template<typename ReturnType>
ReturnType odometer<T>::product_digits() const
{
    BOOST_MPL_ASSERT((::boost::is_convertible<int, ReturnType>));
    BOOST_MPL_ASSERT((::boost::has_multiplies<ReturnType, T>));
    return ::std::accumulate(current_.begin(), current_.end(),
                             static_cast<ReturnType>(1), ::std::multiplies<T>());
}

template<typename T>
typename odometer<T>::const_reference odometer<T>::maximum(size_type n) const
{
    assert(n < maximum_.size());
    return maximum_[n];
}

template<typename T>
typename odometer<T>::reference odometer<T>::maximum(size_type n)
{
    assert(n < maximum_.size());
    return maximum_[n];
}

template<typename T>
typename odometer<T>::const_reference odometer<T>::operator[](size_type n) const
{
    assert(n < current_.size());
    return current_[n];
}

template<typename T>
typename odometer<T>::reference odometer<T>::operator[](size_type n)
{
    assert(n < current_.size());
    return current_[n];
}

template<typename T>
typename odometer<T>::iterator odometer<T>::begin()
{
    return ::boost::make_zip_iterator(::boost::make_tuple(current_.begin(),
                                                          maximum_.begin()));
}

template<typename T>
typename odometer<T>::const_iterator odometer<T>::begin() const
{
    return ::boost::make_zip_iterator(::boost::make_tuple(current_.begin(),
                                                          maximum_.begin()));
}

template<typename T>
typename odometer<T>::const_iterator odometer<T>::cbegin() const
{
    return ::boost::make_zip_iterator(::boost::make_tuple(current_.begin(),
                                                          maximum_.begin()));
}

template<typename T>
typename odometer<T>::iterator odometer<T>::end()
{
    return ::boost::make_zip_iterator(::boost::make_tuple(current_.end(),
                                                          maximum_.end()));
}

template<typename T>
typename odometer<T>::const_iterator odometer<T>::end() const
{
    return ::boost::make_zip_iterator(::boost::make_tuple(current_.end(),
                                                          maximum_.end()));
}

template<typename T>
typename odometer<T>::const_iterator odometer<T>::cend() const
{
    return ::boost::make_zip_iterator(::boost::make_tuple(current_.end(),
                                                          maximum_.end()));
}

template<typename T>
void odometer<T>::reset()
{
    ::std::fill(current_.begin(), current_.end(), T());
    rollover_ = false;
}

template<typename T>
void odometer<T>::reset_to_max()
{
    current_ = maximum_;
    rollover_ = false;
}

template<typename T>
void odometer<T>::clear_rollover()
{
    rollover_ = false;
}

template<typename T>
odometer<T>& odometer<T>::operator++()
{
    BOOST_MPL_ASSERT((::boost::has_pre_increment<T>));
    assert(maximum_.size() == current_.size());

    size_type i = 0;
    while (i < maximum_.size() && ++current_[i] > maximum_[i])
    {
        current_[i] = T();
        ++i;
    }
    if (i == maximum_.size())
    {
        rollover_ = true;
    }
    return *this;
}

template<typename T>
odometer<T> odometer<T>::operator++(int)
{
    odometer<T> tmp(*this);
    ++(*this);
    return tmp;
}

template<typename T>
odometer<T>& odometer<T>::operator--()
{
    BOOST_MPL_ASSERT((::boost::has_pre_decrement<T>));
    assert(maximum_.size() == current_.size());

    size_type i = 0;
    while (i < maximum_.size())
    {
        if (current_[i] == T())
        {
            current_[i] = maximum_[i];
            ++i;
        }
        else
        {
            --current_[i];
            break;
        }
    }
    if (i == maximum_.size())
    {
        rollover_ = true;
    }
    return *this;
}

template<typename T>
odometer<T> odometer<T>::operator--(int)
{
    odometer<T> tmp(*this);
    --(*this);
    return tmp;
}

/// \endcond

} // namespace bnl

#endif // BNL_UTILITY_ODOMETER_HPP
