#include <PrintXDSL.hpp>
#include <graphviz.hpp>
#include <bnl/random/algorithm/random_shuffle.hpp>
#include <bnl/random/randgen.hpp>
#include <bnl/utility/error.hpp>
#include <bnl/algorithm/join.hpp>
#include <bnl/smile_ext/intarray_range.hpp>
#include <bnl/smile_ext/idarray_range.hpp>

#include <boost/assert.hpp>
#include <boost/bind.hpp>
#include <boost/range/adaptor/transformed.hpp>
#include <boost/typeof/typeof.hpp>
#include <boost/unordered_set.hpp>
#include <boost/version.hpp>

#include <cassert>

namespace bnl
{

bayesian_network::bayesian_network()
{
}

bayesian_network::bayesian_network(const ::DSL_network& net)
    : DSL_network(net)
{
}

bayesian_network::bayesian_network(::DSL_network& net, bool fast)
    : DSL_network(net, fast)
{
}

bayesian_network::bayesian_network(int num_nodes, int initial_delta)
    : DSL_network(num_nodes, initial_delta)
{
}

void bayesian_network::randomize_arcs(const ::std::vector<int>& order,
                                      unsigned max_parents,
                                      bool shuffle_possible_parents)
{
    // Make sure the total order can be used on the network
    if (!can_use_order(order))
    {
        throw match_error("Match Error: Input total order doesn't match network.");
    }

    RemoveAllArcs();
    if (shuffle_possible_parents)
    {
        // Attempt to reduce bias by shuffling potential parents before selection
        typedef ::std::vector<int>::iterator iterator;
        ::std::vector<int> order_copy(order);
        const iterator first = order_copy.begin();
        for (iterator i = first; i != order_copy.end(); ++i)
        {
            algorithm::random_shuffle(first, i);
            unsigned num_parents = 0;
            for (iterator j = first; j != i && num_parents < max_parents; ++j)
            {
                if (random::random_bool())
                {
                    BOOST_VERIFY(AddArc(*j, *i) == DSL_OKAY);
                    ++num_parents;
                }
            }
        }
    }
    else
    {
        // A more simple algorithm that reduces memory overhead
        typedef ::std::vector<int>::const_iterator const_iterator;
        const const_iterator first = order.begin();
        for (const_iterator i = first; i != order.end(); ++i)
        {
            unsigned num_parents = 0;
            for (const_iterator j = first; j != i && num_parents < max_parents; ++j)
            {
                if (random::random_bool())
                {
                    BOOST_VERIFY(AddArc(*j, *i) == DSL_OKAY);
                    ++num_parents;
                }
            }
        }
    }
}
void bayesian_network::randomize_arcs(bool use_order, unsigned max_parents,
                                      bool shuffle_possible_parents)
{
    if (use_order)
    { // Randomize arcs using internal total order
        randomize_arcs(total_order_, max_parents, shuffle_possible_parents);
    }
    else if (GetNumberOfNodes() > 1)
    { // Given no order, just add arcs and redo if a cycle exists
        while (!randomize_arcs_without_order());
    }
}

bool bayesian_network::randomize_arcs_without_order()
{
    // Clear any previous arcs
    RemoveAllArcs();

    // Get the first and second node in network
    const int first_node = GetFirstNode();
    const int second_node = GetNextNode(first_node);

    // Iterate over every pair of nodes in the network
    for (int i = second_node; i != DSL_OUT_OF_RANGE; i = GetNextNode(i))
    {
        for (int j = first_node; j != i; j = GetNextNode(j))
        {
            // Randomly choose one of three states between two nodes
            const int choice = random::random_int(0, 2);
            assert(choice >= 0 && choice <= 2);

            // Add the node (return false on cycle)
            switch (choice)
            {
            case 0:
                break;
            case 1:
                if (AddArc(i, j) != DSL_OKAY)
                {
                    return false;
                }
                break;
            case 2:
                if (AddArc(j, i) != DSL_OKAY)
                {
                    return false;
                }
                break;
            }
        }
    }

    // True on successful addition of all arcs
    return true;
}
void bayesian_network::randomize_order()
{
    ::DSL_errorStringHandler& error_handler = ErrorHandler();
    ::DSL_intArray node_array;
    const int* nodes;
    error_handler.Flush();
    if (GetAllNodes(node_array) != DSL_OKAY || !(nodes = node_array.Items()))
    {
        throw smile_error(error_handler,
                          "SMILE Error: Couldn't get the nodes to create order.");
    }
    total_order_.assign(nodes, nodes + node_array.NumItems());
    algorithm::random_shuffle(total_order_);
}
bool bayesian_network::can_use_order(const ::std::vector<int>& order) const
{
    assert(GetNumberOfNodes() >= 0);
    typedef ::std::vector<int>::const_iterator const_iterator;
    typedef ::boost::unordered_set<int>::size_type size_type;

    const size_type num_nodes = static_cast<size_type>(GetNumberOfNodes());
    ::boost::unordered_set<int> visited;
#if BOOST_VERSION >= 105000
    visited.reserve(num_nodes);
#endif
    for (const_iterator first = order.begin(); first != order.end(); ++first)
    {
        if (!Ok(*first) || !visited.insert(*first).second)
        {
            return false;
        }
    }
    return (visited.size() == num_nodes && num_nodes > 0);
}


bool bayesian_network::totally_ordered(const ::std::vector<int>& order) const
{
    typedef ::std::vector<int>::const_iterator const_iterator;
    const const_iterator last = order.end();
    for (const_iterator me = order.begin(); me != last; ++me)
    {
        for (const_iterator he = me; ++he != last;)
        {
            if (IsThisGuyMyParent(*me, *he) != 0)
            {
                return false;
            }
        }
    }
    return can_use_order(order);
}

bool bayesian_network::totally_ordered() const
{
    return totally_ordered(total_order_);
}

const ::std::vector<int>& bayesian_network::total_order() const
{
    return total_order_;
}

const char* bayesian_network::get_id(int handle) const
{
    ::DSL_node* const node = GetNode(handle);
    const char* id;
    return (!node || !(id = node->Info().Header().GetId())) ? "" : id;
}

const char* bayesian_network::get_outcome_name(int handle, int index) const
{
    ::DSL_node* const node = GetNode(handle);
    ::DSL_nodeDefinition* def;
    ::DSL_idArray* outcomes;
    if (!node || !(def = node->Definition()) || !(outcomes = def->GetOutcomesNames()))
    {
        return "";
    }
    else
    {
        return (*outcomes)[index];
    }
}

void bayesian_network::write_file(const ::std::string& filename, int filetype, void* ext)
{
    ::DSL_errorStringHandler& error_handler = ErrorHandler();
    SimpleGraphLayout();
    error_handler.Flush();
    if (WriteFile(filename.c_str(), filetype, ext) != DSL_OKAY)
    {
        throw io_error((error_handler.GetNumberOfErrors() > 0)
                       ? error_handler.GetLastErrorMessage()
                       : "I/O Error: Unable to write network to " + filename);
    }
}
/*
void bayesian_network::print_node_properties(int handle, ::std::ostream& os) const
{
    // Print handle and try to get the corresponding node
    os << "Handle: " << handle << '\n';
    ::DSL_node* const node = GetNode(handle);
    if (!node)
    {
        os << "Error: No such node exists!\n";
        return;
    }

    // Get some node information and print identifier/name
    const ::DSL_header& header = node->Info().Header();
    const char* const id = header.GetId();
    const char* const name = header.GetName();
    if (id)
    {
        os << "Id: " << id << '\n';
    }
    if (name)
    {
        os << "Name: " << name << '\n';
    }

    // Prepare member function to get id as a unary function for later use
    BOOST_AUTO(handle_to_id, ::boost::bind(&bayesian_network::get_id, this, _1));

    // Print the parents (number and identifiers)
    const ::DSL_intArray& parents = GetParents(handle);
    const int num_parents = parents.NumItems();
    os << "Parents[" << num_parents << "]: {";
    algorithm::join(parents | ::boost::adaptors::transformed(handle_to_id), ',', os);
    os << "}\n";

    // Print the children (number and identifiers)
    const ::DSL_intArray& children = GetChildren(handle);
    os << "Children[" << children.NumItems() << "]: {";
    algorithm::join(children | ::boost::adaptors::transformed(handle_to_id), ',', os);
    os << "}\n";

    // Print the outcome (number and names)
    ::DSL_nodeDefinition* const definition = node->Definition();
    const ::DSL_idArray* outcomes_names;
    if (!definition || !(outcomes_names = definition->GetOutcomesNames()))
    {
        return;
    }
    os << "Outcomes[" << outcomes_names->NumItems() << "]: {";
    algorithm::join(*outcomes_names, ',', os);
    os << "}\n";

    // Print conditional probabilities
    ::DSL_sysCoordinates definition_coords(*definition);
    definition_coords.GoFirst();
    os << "Conditional Probabilities:\n";
    do
    {
        ::DSL_intArray& coords = definition_coords.Coordinates();
        os << "  P(" << id << '=' << (*outcomes_names)[coords.Last()];
        if (num_parents > 0)
        {
            os << " | " << get_id(parents[0]) << '=';
            os << get_outcome_name(parents[0], coords[0]);
            for (int i = 1; i < num_parents; ++i)
            {
                os << ", " << get_id(parents[i]) << '=';
                os << get_outcome_name(parents[i], coords[i]);
            }
        }
        os << ") = " << definition_coords.UncheckedValue() << '\n';
    } while (definition_coords.Next() != DSL_OUT_OF_RANGE);
}
*/

void bayesian_network::probabilities(int handle)
{
	// Input conditional probabilities
    ::DSL_node* const node = GetNode(handle);
    if (!node)
    {
        throw io_error("Error: No such node exists!\n");
    }

    const ::DSL_intArray& parents = GetParents(handle);
    const int num_parents = parents.NumItems();

    ::DSL_nodeDefinition* const definition = node->Definition();
    const ::DSL_idArray* outcomes_names;
    if (!definition || !(outcomes_names = definition->GetOutcomesNames()))
    {
        return;
    }

    ::DSL_sysCoordinates definition_coords(*definition);
    definition_coords.GoFirst();
    do
    {
        ::DSL_intArray& coords = definition_coords.Coordinates();
        (*outcomes_names)[coords.Last()];
        if (num_parents > 0)
        {
            get_id(parents[0]);
            get_outcome_name(parents[0], coords[0]);
            for (int i = 1; i < num_parents; ++i)
            {
                get_id(parents[i]);
                get_outcome_name(parents[i], coords[i]);
            }
        }
        definition_coords.UncheckedValue();
        std::cout << definition_coords.UncheckedValue() << std::endl;
        
    } while (definition_coords.Next() != DSL_OUT_OF_RANGE);
}  
/*
::std::ostream& operator<<(::std::ostream& os, const bayesian_network& bn)
{
    // Access the first node handle in the network
    int handle = bn.GetFirstNode();
    if (handle != DSL_OUT_OF_RANGE)
    {
        bn.print_node_properties(handle, os);

        // Loop over every other node and output properties
        while ((handle = bn.GetNextNode(handle)) != DSL_OUT_OF_RANGE)
        {
            bn.print_node_properties(handle, os << '\n');
        }
    }
    return os;
}

bayesian_network add_parameters(refcounted_dataset& dataset,
                                     bool use_order,
                                     unsigned max_parents,
                                     bool shuffle_possible_parents)
{
   
    bayesian_network bn;
    dataset.prepare_structure(bn);
    bn.randomize_order();
    bn.randomize_arcs(use_order, max_parents, shuffle_possible_parents);
    int handle = bn.GetFirstNode();
    if (handle != DSL_OUT_OF_RANGE)
    {
        bn.probabilities(handle);

        // Loop over every other node properties
        while ((handle = bn.GetNextNode(handle)) != DSL_OUT_OF_RANGE)
        {
            bn.probabilities(handle);
        }
    }
   
    
    return bn;
} */
} // namespace bnl
