#include <graphviz.hpp>

#include <boost/algorithm/string/classification.hpp>
#include <boost/algorithm/string/split.hpp>
#include <bnl/utility/error.hpp>
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/graphviz.hpp>
#include <boost/property_map/dynamic_property_map.hpp>
#include <boost/property_map/property_map.hpp>

#include <boost/tuple/tuple.hpp>
#include <boost/unordered_map.hpp>
#include <boost/version.hpp>

namespace bnl
{

void read_graphviz(::std::istream& in, bayesian_network& bn)
{
    using namespace boost;
    typedef property<vertex_name_t, ::std::string,
            property<vertex_color_t, ::std::string> > vertex_p;
    typedef adjacency_list<vecS, vecS, directedS, vertex_p, no_property,
                           property<graph_name_t, ::std::string> > graph_t;

    // Initialize Boost graph and dynamic properties
    graph_t g(0);
    dynamic_properties dp(&ignore_other_properties);
    property_map<graph_t, vertex_name_t>::type name = get(vertex_name, g);
    property_map<graph_t, vertex_color_t>::type outcomes = get(vertex_color, g);
    ref_property_map<graph_t*, ::std::string> gname(get_property(g, graph_name));
    dp.property("label", name);
    dp.property("outcomes", outcomes);
    dp.property("name", gname);

    // Run the Boost.Graph GraphViz parser
    try
    {
        if (!read_graphviz(in, g, dp, "label"))
        {
            throw io_error("I/O Error: Failed to create Boost.Graph from DOT file.");
        }
    }
    catch (const graph_exception& e)
    {
        throw io_error(e.what());
    }

    // Use graph-level properties to alter basic Bayesian network properties
    ::DSL_header& network_header = bn.Header();
    const ::std::string& network_name = gname[&g];
    if (!network_name.empty())
    {
        network_header.SetName(network_name.c_str());
        ::std::vector<char> network_id(network_name.begin(), network_name.end());
        network_id.push_back('\0');
        if (::DSL_header::MakeValidId(&network_id[0]) == DSL_OKAY)
        {
            network_header.SetId(&network_id[0]);
        }
    }

    // Use vertex-level properties to add nodes
    unordered_map< ::std::string, int> name_to_handle;
#if BOOST_VERSION >= 105000
    name_to_handle.reserve(num_vertices(g));
#endif
    bn.DeleteAllNodes();
    graph_traits<graph_t>::vertex_iterator vi, vi_end;
    for (tie(vi, vi_end) = vertices(g); vi != vi_end; ++vi)
    {
        // Create network node corresponding to Boost.Graph vertex
        const ::std::string& node_name = name[*vi];
        int handle;
        if (node_name.empty())
        {
            throw io_error("I/O Error: A node does not have a valid name.");
        }
        else
        {
            ::std::vector<char> node_id;
            node_id.reserve(node_name.size() + 2);
            if (::std::isdigit(node_name[0]))
            {
                node_id.push_back('N');
            }
            node_id.insert(node_id.end(), node_name.begin(), node_name.end());
            node_id.push_back('\0');

            if (::DSL_header::MakeValidId(&node_id[0]) == DSL_OKAY)
            {
                handle = bn.AddNode(DSL_CPT, &node_id[0]);
            }
            else
            {
                handle = bn.AddNode(DSL_CPT, 0);
            }
        }
        if (!bn.Ok(handle))
        {
            throw io_error("I/O Error: Unable to add node to network.");
        }
        name_to_handle[node_name] = handle;

        // Try to parse outcomes names for the node
        DSL_nodeDefinition* const definition = bn.GetNode(handle)->Definition();
        if (!definition)
        {
            throw io_error("I/O Error: A node's definition does not exist.");
        }
        const ::std::string& node_outcomes_str = outcomes[*vi];
        int success_setting_outcomes;
        if (node_outcomes_str.empty())
        {
            success_setting_outcomes = definition->SetNumberOfOutcomes(2);
        }
        else
        {
            ::DSL_idArray node_outcomes;
            ::std::vector< ::std::string> node_outcomes_vec;
            split(node_outcomes_vec, node_outcomes_str, is_any_of(","), token_compress_on);
            ::std::vector< ::std::string>::iterator it;
            for (it = node_outcomes_vec.begin(); it != node_outcomes_vec.end(); ++it)
            {
                ::std::vector<char> node_outcome;
                node_outcome.reserve(it->size() + 2);
                if (!it->empty() && ::std::isdigit((*it)[0]))
                {
                    node_outcome.push_back('S');
                }
                node_outcome.insert(node_outcome.end(), it->begin(), it->end());
                node_outcome.push_back('\0');

                int success_adding_outcome;
                if (::DSL_header::MakeValidId(&node_outcome[0]) == DSL_OKAY)
                {
                    success_adding_outcome = node_outcomes.Add(&node_outcome[0]);
                }
                else
                {
                    success_adding_outcome = node_outcomes.CreateAndAddValidId();
                }
                if (success_adding_outcome != DSL_OKAY)
                {
                    throw io_error("I/O Error: Unable to add outcome name.");
                }
            }
            for (int i = node_outcomes.NumItems(); i < 2; ++i)
            {
                if (node_outcomes.CreateAndAddValidId() != DSL_OKAY)
                {
                    throw io_error("I/O Error: Unable to add outcome name.");
                }
            }
            success_setting_outcomes = definition->SetNumberOfOutcomes(node_outcomes);
        }
        if (success_setting_outcomes != DSL_OKAY)
        {
            throw io_error("I/O Error: Unable to set a node's outcomes names.");
        }
    }

    // Use edge-level properties to add arcs
    graph_traits<graph_t>::edge_iterator ei, ei_end;
    for (tie(ei, ei_end) = edges(g); ei != ei_end; ++ei)
    {
        const int parent = name_to_handle[name[source(*ei, g)]];
        const int child = name_to_handle[name[target(*ei, g)]];
        if (bn.AddArc(parent, child) != DSL_OKAY)
        {
            throw io_error("I/O Error: Unable to add arc between two nodes.");
        }
    }
}

bayesian_network read_graphviz(::std::istream& in)
{
    bayesian_network bn;
    read_graphviz(in, bn);
    return bn;
}

} // namespace bnl
